/*
 * Routines for dealing with the map
 */
#include <sys/types.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_topo_map.h"
#include "lf_fms_comm.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_fms.h"
#include "fma_fabric.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_dfs_route.h"
#include "fma_standalone.h"
#include "fma_probe.h"
#include "fma_verify.h"

/*
 * Local prototypes
 */
static void fma_alloc_route_buffers(struct lf_fabric *fp);
static void fma_clear_routes(struct lf_fabric *fp);
/*
 * Initialize map info
 */
void
fma_init_map_info_vars()
{
  LF_CALLOC(A.map_info, struct fma_map_info, 1);
  return;

 except:
  fma_perror_exit(1);
}


/*
 * We got a new map, translate it into a fabric struct
 * and then do some validation.
 */
int
fma_copy_and_load_map(
  struct lf_topo_map *topo_map_in,
  int topo_map_size)
{
  int rc;
  struct lf_topo_map *topo_map;

  /* Make a copy of the topo map */
  topo_map = (struct lf_topo_map *) malloc(topo_map_size);
  if (topo_map == NULL) LF_ERROR(("Error allocating topo_map buffer"));
  memcpy(topo_map, topo_map_in, topo_map_size);

  /* Load this map struct */
  fma_load_map_struct(topo_map, topo_map_size);

  /* validate the map */
  rc = fma_validate_map();

  /* map is OK, route it */
  if (rc == 0) {
    rc = fma_route_topo_map();

    /* print any errors */
    if (rc != 0) {
      fma_perror();
    }

  /* map is not valid */
  } else {
    fma_map_is_invalid(TRUE, "validation failed");
  }
  return rc;

 except:
  fma_perror_exit(1);
  return -1;
}

/*
 * Release the current topo map
 */
void
fma_release_topo_map()
{
  /* Cancel any map distribution in progress since the map is going away */
  fma_cancel_map_distribution();

  LF_FREE(A.map_info->current_topo_map);
}

/*
 * Turn a map file into a fabric struct
 */
void
fma_load_map_struct(
  struct lf_topo_map *topo_map,
  int topo_map_size)
{
  struct lf_fabric *fp;
  int rc;

  /* Release the current topo map */
  fma_release_topo_map();

  /* Set the new topo map */
  A.map_info->current_topo_map = topo_map;
  A.map_info->current_topo_map_size = topo_map_size;

  /* get pointer to map holder */
  lf_free_fabric(A.fabric);
  fp = fma_create_fabric();
  A.fabric = fp;

  fma_log("  %d hosts, %d nics, %d xbars, %d links",
    ntohl(topo_map->num_hosts_32), ntohl(topo_map->num_nics_32),
    ntohl(topo_map->num_xbars_32), ntohl(topo_map->num_links_32));
  fma_log("mapper is " LF_MAC_FORMAT ", version is %d",
      LF_MAC_ARGS(topo_map->mapper_mac_addr), ntohl(topo_map->map_id_32));

  /* save mapper info */
  LF_MAC_COPY(A.map_info->mapper_mac_addr, topo_map->mapper_mac_addr);
  A.map_info->mi_map_version = ntohl(topo_map->map_id_32);

  /* copy over counts for xbars and NICs */
  fp->num_hosts = ntohl(topo_map->num_hosts_32);
  fp->num_xbars = ntohl(topo_map->num_xbars_32);
  FMA_FABRIC(fp)->num_nics = ntohl(topo_map->num_nics_32);
  FMA_FABRIC(fp)->num_links = ntohl(topo_map->num_links_32);

  /* allocate space for xbars and NICs */
  LF_CALLOC(fp->hosts, struct lf_host *, fp->num_hosts);
  LF_CALLOC(fp->xbars, struct lf_xbar *, fp->num_xbars);
  LF_CALLOC(FMA_FABRIC(fp)->nics, struct lf_nic *, FMA_FABRIC(fp)->num_nics);
  LF_CALLOC(FMA_FABRIC(fp)->link_end_node, union lf_node *,
            FMA_FABRIC(fp)->num_links);
  LF_CALLOC(FMA_FABRIC(fp)->link_end_port, char, FMA_FABRIC(fp)->num_links);

  /* Allocate all the hosts */
  {
    struct lf_host *hp;
    int h;

    for (h=0; h<fp->num_hosts; ++h) {
      hp = fma_create_host();
      if (hp == NULL) LF_ERROR((("Error allocating host")));
      fp->hosts[h] = hp;

      FMA_HOST(hp)->host_id = h;	/* assign host index */
    }
  }

  /* copy over all the xbars */
  {
    struct lf_topo_xbar *topo_xbar;
    int x;

    topo_xbar = LF_TOPO_XBAR_ARRAY(topo_map);

    for (x=0; x<fp->num_xbars; ++x) {
      struct lf_xbar *xp;

      xp = fma_create_xbar(ntohs(topo_xbar->num_ports_16),
	                   ntohl(topo_xbar->xbar_id_32),
			   topo_xbar->quadrant_disable_8);
      if (xp == NULL) LF_ERROR(("Creating xbar"));

      fp->xbars[x] = xp;

      /* fill in stuff from map */
      xp->clos_level = topo_xbar->clos_level_8;
      FMA_XBAR(xp)->node_id = ntohl(topo_xbar->node_id_32);
      xp->x_topo_index = x;

      ++topo_xbar;
    }
  }
  
  /* copy over all the NICs */
  {
    struct lf_topo_nic *topo_nic;
    int nic_ports;
    int n;
    int h;

    topo_nic = LF_TOPO_NIC_ARRAY(topo_map);
    nic_ports = 0;		/* remember most ports on any NIC */

    for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
      struct lf_nic *nicp;
      struct lf_host *hp;

      nicp = fma_create_nic(topo_nic->num_ports_8, topo_nic->mac_addr,
	  topo_nic->firmware_type_8);
      if (nicp == NULL) LF_ERROR(("Creating nic"));

      FMA_FABRIC(fp)->nics[n] = nicp;
      nicp->n_topo_index = n;
      FMA_NIC(nicp)->node_id = ntohl(topo_nic->node_id_32);

      /* use node ID to get index of host */
      h = LF_TOPO_NIC_HOST_ID(FMA_NIC(nicp)->node_id);
      hp = fp->hosts[h];
      rc = lf_add_existing_nic_to_host(nicp, hp);
      if (rc != 0) LF_ERROR((("Error adding NIC to host")));

      nicp->host_nic_id = LF_TOPO_NIC_HOST_NIC_ID(FMA_NIC(nicp)->node_id);
      hp->fma_flags = ntohs(topo_nic->flags_16);
      hp->fw_type = topo_nic->firmware_type_8;
      nicp->partition = ntohs(topo_nic->partition_16);
      nicp->mag_id = ntohs(topo_nic->mag_id_16);

      if (A.debug > 1) {
	fma_log("  " LF_MAC_FORMAT ", flags=%x, mag_id=%d",
	    LF_MAC_ARGS(nicp->mac_addr), hp->fma_flags, nicp->mag_id);
      }

      /* Make sure "max_mac_addr" is up-to-date for owning host */
      if (LF_MAC_CMP(nicp->mac_addr, FMA_HOST(hp)->max_mac_addr) > 0) {
	LF_MAC_COPY( FMA_HOST(hp)->max_mac_addr, nicp->mac_addr);

	/* Look for us in the map */
	if (LF_MAC_CMP(A.myri->my_max_mac, nicp->mac_addr) == 0) {
	  A.my_host = hp;
	}
      }

      if (nicp->num_ports > nic_ports) {
	nic_ports = nicp->num_ports;
      }

      ++topo_nic;
    }

    FMA_FABRIC(fp)->max_nic_ports = nic_ports;	/* for allocating route space */
  }

  /* allocate space for routes */
  fma_alloc_route_buffers(fp);

  /* Now, make all the links */
  {
    struct lf_topo_link *topo_link;
    lf_topo_link_state_t *topo_link_state;
    int which;
    int vni;
    int l;

    topo_link = LF_TOPO_LINK_ARRAY(topo_map);
    topo_link_state = LF_TOPO_LINK_STATE_ARRAY(topo_map);

    for (l=0; l<FMA_FABRIC(fp)->num_links; ++l) {
      union lf_node *np[2];
      int port[2];
      int i;

      /* get pointers to the two nodes */
      for  (i=0; i<2; ++i) {
	int node_index;

	node_index = ntohl(topo_link->node_index_32[i]);

	if (LF_TOPO_NI_XBAR(node_index)) {
	  np[i] = LF_NODE(fp->xbars[LF_TOPO_NI_INDEX(node_index)]);
	} else {
	  np[i] = LF_NODE(FMA_FABRIC(fp)->nics[LF_TOPO_NI_INDEX(node_index)]);
	}
	port[i] = LF_TOPO_NI_PORT(node_index);
      }

      /* get verifier */
      vni = ntohl(topo_link->verifier_32);
      which = LF_TOPO_NI_WHICH_END(vni);

      /* link them to each other */
      for  (i=0; i<2; ++i) {
	if (np[i]->ln_type == LF_NODE_XBAR) {
	  struct lf_xbar *xp;
	  xp = LF_XBAR(np[i]);

	  xp->topo_ports[port[i]] = np[1-i];
	  xp->topo_rports[port[i]] = port[1-i];

	  /* If there is a verifier and this is the right end, save the info */
	  if (vni != 0 && which == i) {
	    xp->verifiers[port[i]].ver_nicp =
	      FMA_FABRIC(fp)->nics[LF_TOPO_NI_INDEX(vni)];
	    xp->verifiers[port[i]].ver_port = LF_TOPO_NI_PORT(vni);
	  }

	} else {
	  struct lf_nic *nicp;
	  nicp = LF_NIC(np[i]);

	  nicp->topo_ports[port[i]] = np[1-i];
	  nicp->topo_rports[port[i]] = port[1-i];
	}
      }

      /* set the link state */
      FMA_FABRIC(fp)->link_end_node[l] = np[0];
      FMA_FABRIC(fp)->link_end_port[l] = port[0];
      lf_set_node_link_state(np[0], port[0], *topo_link_state);

      ++topo_link;
      ++topo_link_state;
    }
  }

  /* Truncate "num_ports" on each xbar to make everything go a little faster */
  {
    int x;
    int p;

    for (x=0; x<fp->num_xbars; ++x) {
      struct lf_xbar *xp;
      xp = fp->xbars[x];
      for (p=xp->num_ports-1; p>0; --p) {
	if (xp->topo_ports[p] != NULL) {
	  break;
	}
      }
      xp->num_ports = p+1;
    }
  }

  /* calculate the clos_level for each xbar */
  rc = lf_clos_label_xbars(A.fabric);
  if (rc == -1) LF_ERROR((("Error getting clos levels for xbars")));

  fma_find_nic_indices();	/* update NIC indices */
  fma_calc_xbar_dists();	/* calc distance from each xbar to this host */

  return;

 except:
  fma_perror_exit(1);
}

/*
 * We got a new link state from the FMS, update our map with the new states
 */
void
fma_load_link_state(
  lf_topo_link_state_t *topo_link_state)
{
  struct lf_fabric *fp;
  lf_topo_link_state_t *lsp;
  int l;

  /* get pointer to fabric */
  fp = A.fabric;

  /* Copy this new link state into our topo map */
  lsp = LF_TOPO_LINK_STATE_ARRAY(A.map_info->current_topo_map);
  memcpy(lsp, topo_link_state, FMA_FABRIC(fp)->num_links);

  /* set the link state for each link */
  for (l=0; l<FMA_FABRIC(fp)->num_links; ++l) {

    lf_set_node_link_state(FMA_FABRIC(fp)->link_end_node[l],
		           FMA_FABRIC(fp)->link_end_port[l],
		           topo_link_state[l]);
  }

  /* xbar distances may have changed */
  fma_calc_xbar_dists();	/* calc distance from each xbar to this host */
}

/*
 * Allocate all route buffers
 */
static void
fma_alloc_route_buffers(
  struct lf_fabric *fp)
{
  int n;
  int size;
  int num_lens;
  struct lf_nic *nicp;

  num_lens =  FMA_FABRIC(fp)->max_nic_ports * FMA_FABRIC(fp)->max_nic_ports * A.myri->num_nics *
    A.myri->nic_info[0]->myri_info.num_routes;      
  size = num_lens * FMA_IFC_ROUTE_LEN;

  for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
    nicp = FMA_FABRIC(fp)->nics[n];
    LF_CALLOC(FMA_NIC(nicp)->route_buf, unsigned char, size);
    LF_CALLOC(FMA_NIC(nicp)->route_lens, int8_t, num_lens);
  }

  /* routes start cleared */
  fma_clear_routes(fp);

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Clear all routes
 */
static void
fma_clear_routes(
  struct lf_fabric *fp)
{
  int n;
  int num_lens;
  struct lf_nic *nicp;

  num_lens =  FMA_FABRIC(fp)->max_nic_ports * FMA_FABRIC(fp)->max_nic_ports
    * A.myri->num_nics * A.myri->nic_info[0]->myri_info.num_routes;      

  for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
    nicp = FMA_FABRIC(fp)->nics[n];
    memset(FMA_NIC(nicp)->route_lens, -1, num_lens);
  }
}


/*
 * Find a NIC index given mac address
 */
int
fma_mac_to_nic_index(
  lf_mac_addr_t mac)
{
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;
  for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
    if (memcmp(mac, FMA_FABRIC(fp)->nics[n]->mac_addr, sizeof(lf_mac_addr_t)) == 0) {
      return n;
    }
  }
  return -1;
}

/*
 * Find our NICs in this map
 */
void
fma_find_nic_indices()
{
  struct fma_myri *mp;
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;
  mp = A.myri;

  /* find each NIC in map */
  for (n=0; n<mp->num_nics; ++n) {
    struct fma_nic_info *nip;
    
    nip = mp->nic_info[n];
    nip->ni_map_index = fma_mac_to_nic_index(nip->myri_info.mac_addr);

    if (nip->ni_map_index == -1) {
      fma_log("Cannot find NIC %d (" LF_MAC_FORMAT ") in map!",
	nip->nic_id, LF_MAC_ARGS(nip->myri_info.mac_addr));
      nip->nic_ptr = NULL;
    } else {

      /* mark NIC as local to this host */
      FMA_NIC(FMA_FABRIC(fp)->nics[nip->ni_map_index])->local_nip = nip;

     fma_log("Found NIC %d at index %d!", nip->nic_id, nip->ni_map_index);
      nip->nic_ptr = FMA_FABRIC(A.fabric)->nics[nip->ni_map_index];
    }
  }
}

/*
 * Generate routes for the current topo map
 */
int
fma_route_topo_map()
{
  int num_bad;

  /* Validate our quick routes.  Re-calculate if any are bad */
  num_bad = fma_check_quick_routes();
  if (A.debug) fma_log("%d quick routes need to be re-calculated", num_bad);

  /* recalculate quick_routes if needed */
  if (num_bad > 0) {
    fma_calc_all_quick_routes();
  }

  /* clear current routes */
  fma_clear_routes(A.fabric);

  /* perform real route calculation. */
  fma_dfs_calc_all_routes(A.myri->nic_info[0]->myri_info.num_routes);

  /* load the routes into firmware */
  fma_set_all_routes();

  return 0;
}

/*
 * validate map
 */
int
fma_validate_map()
{
  struct fma_myri *fmp;
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;

  fmp = A.myri;
  for (n=0; n<fmp->num_nics; ++n) {
    struct fma_nic_info *nip;
    struct lf_nic *nicp;
    int p;

    nip = fmp->nic_info[n];

    /* Is this NIC even in the map? */
    if (nip->ni_map_index == -1) {

      /* not in map - if connected, map is invalid */
      if (nip->xbar_id[0] != -1) {
	fma_log("Map does not see my connected NIC");
	return -1;
      }

    /* found NIC in map - compare connectedness */
    } else {

      nicp = FMA_FABRIC(fp)->nics[nip->ni_map_index];
      for (p=0; p<nicp->num_ports; ++p) {
	struct lf_xbar *xp;
	int xport;

	xp = LF_XBAR(nicp->topo_ports[p]);
	xport = nicp->topo_rports[p];

	/* if we don't believe we are disconnected but map does,
	 * then call it invalid
	 */
	if (nip->xbar_id[p] != -1 && xp == NULL) {
	 fma_log("map shows disconnected NIC %d, port %d", n, p);
	  return -1;
	}

	/* make sure we have a running verify on direct xbar */
	if (xp == NULL || xp->ln_type == LF_NODE_XBAR) {
	  fma_define_direct_xbar_verify_probe(nip, p);
	}
      }
    }
  }

 fma_log("map seems OK");
  return 0;
}

struct lf_host *
fma_find_host_by_max_mac_addr(
  lf_mac_addr_t max_mac_addr)
{
  struct lf_fabric *fp;
  int h;

  fp = A.fabric;

  /* if no map, then definitely no host to be found */
  if (fp == NULL) return NULL;

  /* look through all hosts for a match */
  for (h=0; h<fp->num_hosts; ++h) {
    struct lf_host *hp;

    hp = fp->hosts[h];
    if (LF_MAC_CMP(FMA_HOST(hp)->max_mac_addr, max_mac_addr) == 0) {
      return hp;
    }
  }
  return NULL;
}

/*
 * Map is invalid - call the right routine
 */
void
fma_map_is_invalid(
  int use_me,
  char *why)
{
  if (A.debug) {
    fma_log("Map version %d invalid: %s", A.map_info->mi_map_version, why);
  }

  if (A.run_state == FMA_RUN_STANDALONE) {
    fma_standalone_map_is_invalid();
  } else {
    fma_fms_map_is_invalid(A.map_info->mi_map_version, use_me, why);
  }
}

/*
 * Set our current mapping level
 */
void
fma_set_mapping_level(
  int level)
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;
  sdp->my_level = level;

  /* set the level of all probes */
  fma_set_probe_levels(level);
}

/*
 * Got a remap request from another FMA.  If we're an active mapper, consider
 * the request.  If the map version matches our current map version and map
 * grace timer is in effect, clear the timer and reset map in progress.
 */
void
fma_peer_remap_request(
  struct fma_map_request *reqp)
{
  struct fma_standalone_data *sdp;
  int bad_map_version;

  sdp = A.stand;

  if (A.run_state == FMA_RUN_STANDALONE) {
    if (A.stand->active) {
      if (A.debug) fma_log("Following peer request to map fabric");

      /* If map version matches our current version and we are not really
       * actively mapping, force a remap.
       */
      bad_map_version = ntohl(reqp->map_version_32);
      if (bad_map_version == A.map_info->mi_map_version
	  && A.myri->map_progress == NULL) {
	if (A.debug) {
	  fma_log("Current map version noted invalid, force remap");
	}

	/* Cancel any map distribution */
	fma_cancel_map_distribution();
	fma_reset_mapping_done();
      }

      fma_start_mapping_fabric(fma_standalone_map_done);
    } else {
      if (A.debug) {
	fma_log("Ignoring peer request to map fabric: not active");
      }
    }
  } else if (A.debug) {
    fma_log("Ignoring map request from peer while not standalone");
  }
}
