/*
 * Routines for running in standalone mode
 */
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_fabric.h"
#include "lf_topo_map.h"
#include "lf_fma_flags.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_fabric.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_fms.h"
#include "fma_standalone.h"
#include "fma_tunnel.h"
#include "fma_verify.h"
#include "fma_proxy.h"

/*
 * Local prototypes
 */
static void fma_standalone_map_timeout(void *v);
static void fma_standalone_map_request_timeout(void *v);
static void fma_standalone_dist_map_done(void *v);
static void fma_standalone_dist_map_failed(void *v);
static void fma_request_remap_from_mapper(void);
static void fma_copy_fabric_stuff_to_topo(void);
#if 0
static int fma_is_link_xbar_xbar(int topo_index);
#endif
static void fma_standalone_assign_probes(void);
static int fma_compare_levels( int level1, lf_mac_addr_t mac_addr1,
    int level2, lf_mac_addr_t mac_addr2);
static int fma_standalone_next_map_send(void);
static void fma_map_distribution_complete(void);
static void fma_mapping_grace_expired(void *v);

/*
 * Initialize static things that depend on nothing else
 */
void
fma_init_standalone_vars()
{
  struct fma_standalone_data *sdp;

  LF_CALLOC(A.stand, struct fma_standalone_data, 1);
  sdp = A.stand;

  /* some static message hearders */
  sdp->topo_map_hdr.type_16 = htons(FMA_PACKET_TYPE);
  sdp->topo_map_hdr.subtype_16 = htons(FMA_SUBTYPE_TOPO_MAP);
  return;

 except:
  fma_perror_exit(1);
}

void
fma_init_standalone()
{
}

void
fma_enter_standalone_mode()
{
  struct fma_standalone_data *sdp;
  struct fma_myri *mp;

  if (A.debug) fma_log("Enter standalone mode");

  sdp = A.stand;
  mp = A.myri;

  /* indicate we are in standalone mode */
  A.run_state = FMA_RUN_STANDALONE;
  A.fma_flags &= ~FMA_FLAG_HAS_FMS;

  /* Active by default */
  sdp->active = TRUE;

  /* set our level */
  fma_set_mapping_level(A.map_level);

  sdp->winner_level = sdp->my_level;
  LF_MAC_COPY(sdp->winner_mac, mp->my_max_mac);

  /* Start mapping, cleanest thing to do */
  fma_start_mapping_fabric(fma_standalone_map_done);
}

/*
 * This is called whenever we are entering a new run state from standalone
 */
void
fma_cancel_standalone_mode()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  /* Cancel any mapping that might be in progress */
  fma_mf_cancel_mapping();

  /* if we are waiting on a map, stop waiting now */
  if (sdp->map_wait_timer != NULL) {
    lf_remove_event(sdp->map_wait_timer);
    sdp->map_wait_timer = NULL;		/* indicate the timer has expired */
  }
}

/*
 * We got contact from another NIC out there.
 */
void
fma_standalone_got_contact(
  int level,
  lf_mac_addr_t max_mac_addr,
  struct fma_nic_info *nip,
  int port,
  uint8_t *route,
  int route_len,
  int is_verifying)
{
  struct fma_standalone_data *sdp;
  int map_timeout;
  int cmp;

  sdp = A.stand;

  /* If this mapper if higher than highest so far, then make this the
   * new mapper */
  cmp = fma_compare_levels(level, max_mac_addr,
      			   sdp->winner_level, sdp->winner_mac);
  if (cmp > 0) {

    fma_log("new mapper is " LF_MAC_FORMAT ", level %d",
	LF_MAC_ARGS(max_mac_addr), level);

    /* Save this as address of new mapping winner */
    sdp->winner_level = level;
    LF_MAC_COPY(sdp->winner_mac, max_mac_addr);

    /* remember how to get to mapper */
    sdp->mapper_route_nic = nip;
    sdp->mapper_route_port = port;
    sdp->mapper_route_len = route_len;
    memcpy(sdp->mapper_route, route, route_len);

    /* If we used to be the active, cancel any mapping in progress and 
     * start waiting for a map */
    if (sdp->active) {
      sdp->active = FALSE;
      fma_mf_cancel_mapping();
    }

    /* remove any possible pending timeout events */
    if (sdp->map_wait_timer != NULL) {
      lf_remove_event(sdp->map_wait_timer);
      sdp->map_wait_timer = NULL;
    }

    /* Start waiting for a map from this new mapper */
    if (level > FMA_MAX_MAPPER_LEVEL) {
      map_timeout = FMA_FMS_MAP_WAIT_TIMEOUT;
    } else {
      map_timeout = FMA_STANDALONE_MAP_WAIT_TIMEOUT;
    }
    sdp->map_wait_timer = lf_schedule_event(fma_standalone_map_timeout, NULL,
		                            map_timeout);
    if (sdp->map_wait_timer == NULL) {
      LF_ERROR((("Error scheduling map wait timeout")));
    }

    /*
     * If this mapper has FMA contact, and we don't, use him as a proxy
     * to start communication with the FMS
     */
    if (level >= FMA_FMS_MAPPER_LEVEL) {
      fma_proxy_initiate(nip, port, route, route_len);
    }

  /* not higher, but if this is a new NIC or a NIC not yet
   * in verify mode, cause a remap */
  } else if (cmp < 0 
             && (!is_verifying
                 || (fma_find_host_by_max_mac_addr(max_mac_addr) == NULL))) {
    lf_string_t why;

    if (!sdp->active) {
      cmp = fma_compare_levels(level, max_mac_addr,
	  sdp->my_level, A.myri->my_max_mac);
      if (cmp < 0 && sdp->map_wait_timer == NULL) {
	fma_log("Scouted by lesser mapper - becoming active");
	sdp->active = TRUE;
	LF_MAC_COPY(sdp->winner_mac, A.myri->my_max_mac);
	sdp->winner_level = sdp->my_level;

	if (sdp->map_wait_timer != NULL) {
	  lf_remove_event(sdp->map_wait_timer);
	  sdp->map_wait_timer = NULL;
	}
      }
    }

    /* If not mapping, then this map is invalid */
    if (!sdp->mapping_in_progress) {
      sprintf(why, "scouted by "LF_MAC_FORMAT", lev=%d, verifying=%d",
	  LF_MAC_ARGS(max_mac_addr), level, is_verifying);
      fma_map_is_invalid(TRUE, why);
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * The map wait timeout expired - become the active mapper again and start
 * mapping!
 */
static void
fma_standalone_map_timeout(
  void *v)
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  sdp->map_wait_timer = NULL;		/* indicate the timer has expired */

  fma_log("timeout waiting for map, going active again");

  /* go back to being the active mapper */
  sdp->active = TRUE;
  LF_MAC_COPY(sdp->winner_mac, A.myri->my_max_mac);
  sdp->winner_level = sdp->my_level;


  fma_start_mapping_fabric(fma_standalone_map_done);
}

/*
 * Our existing map is invalid, we need a new one.
 * If we are the active mapper, start mapping, else send a request
 * to the active mapper to re-map
 */
void
fma_standalone_map_is_invalid()
{
  struct fma_standalone_data *sdp;
  int map_timeout;

  sdp = A.stand;

  /* if we're the mapper, map it! */
  if (sdp->active) {
    fma_start_mapping_fabric(fma_standalone_map_done);

  /* otherwise, if we're not already waiting for a map, request one */
  } else if (sdp->map_wait_timer == NULL) {
    fma_request_remap_from_mapper();

    /* Start waiting for a map from this mapper */
    if (sdp->winner_level > FMA_MAX_MAPPER_LEVEL) {
      map_timeout = FMA_FMS_MAP_WAIT_TIMEOUT;
    } else {
      map_timeout = FMA_STANDALONE_MAP_WAIT_TIMEOUT;
    }
    sdp->map_wait_timer = lf_schedule_event(fma_standalone_map_timeout, NULL,
		                            map_timeout);
    if (sdp->map_wait_timer == NULL) {
      LF_ERROR((("Error scheduling map wait timeout")));
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Request a re-map from another standalone mapper
 */
static void
fma_request_remap_from_mapper()
{
  struct fma_standalone_data *sdp;
  struct fma_map_info *mip;
  struct fma_map_request req;

  sdp = A.stand;
  mip = A.map_info;

  /* fill in map request packet */
  req.hdr.type_16 = htons(FMA_PACKET_TYPE);
  req.hdr.subtype_16 = htons(FMA_SUBTYPE_MAP_FABRIC);
  req.map_version_32 = htonl(A.map_info->mi_map_version);

  fma_log("Requesting remap from " LF_MAC_FORMAT,
      LF_MAC_ARGS(mip->mapper_mac_addr));

  sdp->map_request_msg_id = fma_tunnel_send(
        &req, sizeof(req), NULL, 0,
	sdp->mapper_route, sdp->mapper_route_len,
	sdp->mapper_route_nic, sdp->mapper_route_port,
	NULL, fma_standalone_map_request_timeout, NULL);
}

/*
 * Re-map request timed out - same as if map timed out
 */
static void
fma_standalone_map_request_timeout(
  void *v)
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  if (A.debug) fma_log("Remap request timed out");
  sdp->map_request_msg_id = 0;

  if (sdp->map_wait_timer != NULL) {
    lf_remove_event(sdp->map_wait_timer);
    sdp->map_wait_timer = NULL;
  }

  /* Same as if remap itself timed out */
  fma_standalone_map_timeout(NULL);
}

/*
 * we got a map, clear the timer, make probe assignments
 */
void
fma_standalone_load_map(
  struct lf_topo_map *topo_map,
  int topo_map_size)
{
  struct fma_standalone_data *sdp;
  int rc;

  sdp = A.stand;

  /* Make a copy of this map and load it */
  rc = fma_copy_and_load_map(topo_map, topo_map_size);

  /* If map is valid, continue processing it */
  if (rc == 0) {

    /* Cancel remap request if in progress */
    if (sdp->map_request_msg_id != 0) {
      if (A.debug) fma_log("Cancelling remap request");
      fma_tunnel_cancel_send(sdp->map_request_msg_id);
      sdp->map_request_msg_id = 0;
    }

    /* Clear map receipt timeout timer */
    if (sdp->map_wait_timer != NULL) {
      lf_remove_event(sdp->map_wait_timer);
      sdp->map_wait_timer = NULL;
    }

    /* clear old probe assignments, make new ones */
    fma_standalone_assign_probes();
  }
}

/*
 * Compare a given level/max_mac to the current mapper and return <0, 0, >0
 * a la memcmp()
 */
static int
fma_compare_levels(
  int level1,
  lf_mac_addr_t mac_addr1,
  int level2,
  lf_mac_addr_t mac_addr2)
{
  int cmp;

  cmp = level1 - level2;
  if (cmp != 0) return cmp;

  return LF_MAC_CMP(mac_addr1, mac_addr2);
}

/*
 * We have completed a map in standalone mode.  Start sending it to
 * everyone in it.
 */
void
fma_standalone_map_done(
  struct lf_topo_map *topo_map,
  int topo_map_size)
{
  struct fma_myri *fmp;
  struct fma_standalone_data *sp;
  int rc;

  if (A.debug) fma_log("loading standalone map");

  /* load pointers */
  fmp = A.myri;
  sp = A.stand;

  /* load the map we generated */
  fma_load_map_struct(topo_map, topo_map_size);

  /* Create link verify assignments */
  lf_assign_link_verifies(A.fabric);

  /* copy verify assignments and clos levels into the topo map */
  fma_copy_fabric_stuff_to_topo();

  /* generate and load routes from this map */
  rc = fma_route_topo_map();
  if (rc != 0) {
    LF_ERROR(("Error routing from self-generated map"));
  }

  /* Send the topo map to everyone else who needs it */
  fma_distribute_topo_map();

  /* Make probes assignments */
  fma_standalone_assign_probes();

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Distribute this topo map to all other hosts in the topo map.
 * Distribute only once to each host, even where hosts have multiple NICs
 */
void
fma_distribute_topo_map()
{
  struct fma_standalone_data *sdp;
  struct lf_fabric *fp;
  int sent;

  sdp = A.stand;
  fp = A.fabric;			/* our fabric */

  fma_log("Distributing topo map to hosts");

  /* initialize counters for map sending */
  sdp->map_sends_pending = 0;
  sdp->next_map_send = 0;

  do {
    sent = fma_standalone_next_map_send();
  } while (sent > 0);

  /* If nothing distributed, then we are done already! */
  if (sdp->map_sends_pending <= 0) {
    fma_map_distribution_complete();
  }
}

/*
 * post-mapping grace period has expired
 */
static void
fma_mapping_grace_expired(
  void *v)
{
  A.stand->map_grace_timer = NULL;
  fma_reset_mapping_done();
}

/*
 * Called when time to reset mapping in progress
 */
void
fma_reset_mapping_done()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  /* clear grace timer if active */
  if (sdp->map_grace_timer != NULL) {
    lf_remove_event(sdp->map_grace_timer);
    sdp->map_grace_timer = NULL;
  }
  sdp->mapping_in_progress = FALSE;

  if (A.debug) fma_log("Resetting map_in_progress");
}

/*
 * Called when map distrobution is complete
 */
static void
fma_map_distribution_complete()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  fma_log("map distribution complete");

  sdp->map_grace_timer = lf_schedule_event(fma_mapping_grace_expired, NULL, 
      FMA_MAP_GRACE_PERIOD);
  if (sdp->map_grace_timer == NULL) {
    LF_ERROR(("Error starting mapping grace period timer"));
  }
  return;

 except:
  fma_perror_exit(1);
}

static int
fma_standalone_next_map_send()
{
  struct lf_fabric *fp;
  struct lf_host *hp;
  struct fma_host *fhp;
  struct fma_standalone_data *sdp;
  int h;

  sdp = A.stand;
  fp = A.fabric;			/* our fabric */

  /* If no more pendings allowed, do not send */
  if (sdp->map_sends_pending >= FMA_MAX_MAP_SENDS_PENDING) {
    return 0;
  }

  /* If no more hosts, all done */
  h = sdp->next_map_send;
  if (h >= fp->num_hosts) {
    return 0;
  }

  /* Next host to which to send */
  hp = fp->hosts[h];

  /* Skip self, anyone who cannot route, and anyone who has FMS */
  while (hp == A.my_host
	 || FMA_HOST(hp)->quick_route_len < 0
	 || (hp->fma_flags & (FMA_FLAG_CAN_ROUTE|FMA_FLAG_HAS_FMS))
	      != FMA_FLAG_CAN_ROUTE) {
    ++h;

    /* If none left, all done! */
    if (h >= fp->num_hosts) {
      return 0;
    }

    /* get next host */
    hp = fp->hosts[h];
  }

  /* Send it */
  fhp = FMA_HOST(hp);

  if (A.debug) fma_log("Start map dist to %d", h);

  /* Send a map to this host */
  FMA_HOST(hp)->map_send_id = fma_tunnel_send(
	&sdp->topo_map_hdr, sizeof(sdp->topo_map_hdr),
	A.map_info->current_topo_map, A.map_info->current_topo_map_size,
	fhp->quick_route, fhp->quick_route_len,
	fhp->quick_nic, fhp->quick_port,
	fma_standalone_dist_map_done,
	fma_standalone_dist_map_failed,
	(void *) ((uintptr_t)h));

  /* increment pending count */
  ++sdp->map_sends_pending;
  sdp->next_map_send = h+1;

  return 1;		/* indicate we sent something */
}

static void
fma_standalone_dist_map_done(
  void *v)
{
  struct fma_standalone_data *sp;
  struct lf_host *hp;
  int hid;

  sp = A.stand;

  hid = (int) ((uintptr_t)v);
  hp = A.fabric->hosts[hid];

  if (A.debug) {
    fma_log("Map distribution to %d complete, MAC=" LF_MAC_FORMAT,
	hid, LF_MAC_ARGS(hp->nics[0]->mac_addr));
  }

  --sp->map_sends_pending;
  FMA_HOST(hp)->map_send_id = 0;

  /* send another if we can */
  (void) fma_standalone_next_map_send();

  /* If the map has been successfully sent to everyone, free it */
  if (sp->map_sends_pending <= 0) {
    fma_map_distribution_complete();
  }
}

static void
fma_standalone_dist_map_failed(
  void *v)
{
  struct fma_standalone_data *sp;
  struct lf_host *hp;
  int hid;

  hid = (int) ((uintptr_t)v);
  hp = A.fabric->hosts[hid];

  sp = A.stand;

  fma_log("Map distribution to %d failed? MAC=" LF_MAC_FORMAT, hid, LF_MAC_ARGS(hp->nics[0]->mac_addr));

  --sp->map_sends_pending;
  FMA_HOST(hp)->map_send_id = 0;

  /* send another if we can */
  (void) fma_standalone_next_map_send();

  /* If the map has been successfully sent to everyone, free it */
  if (sp->map_sends_pending <= 0) {
    fma_map_distribution_complete();
  }
}

/*
 * Cancel any map distributions in progress
 */
void
fma_cancel_map_distribution()
{
  struct lf_fabric *fp;
  struct lf_host *hp;
  int h;

  fp = A.fabric;

  /* If no fabric, nothing to distribute */
  if (fp == NULL) return;

  /* cancel any map sends in progress on each host */
  for (h=0; h<fp->num_hosts; ++h) {
    hp = fp->hosts[h];
    if (FMA_HOST(hp)->map_send_id != 0) {
      fma_tunnel_cancel_send(FMA_HOST(hp)->map_send_id);
      FMA_HOST(hp)->map_send_id = 0;
    }
  }
}


/*
 * Make probe assignments
 */
void
fma_standalone_assign_probes()
{
  struct lf_fabric *fp;
  struct lf_host*first_verifier;
  struct lf_host*next_verifier;
  int verifiers;
  int my_id;
  int mod;
  int h;
  int x;

  fp = A.fabric;

  /* first, clear all old probes */
  fma_clear_verify_probes();

#if 1
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;
    int p;

    xp = fp->xbars[x];
    for (p=0; p<xp->num_ports; ++p) {
      struct lf_nic *nicp;

      nicp = xp->verifiers[p].ver_nicp;
      if (nicp != NULL
	  && nicp != LF_VERIFY_OTHER_END
	  && nicp->host == A.my_host) {
	if (A.debug > 1) {
	  fma_log("x%dp%d assigned to " LF_MAC_FORMAT " p%d", x, p,
	      LF_MAC_ARGS(nicp->mac_addr), xp->verifiers[p].ver_port);
	}
	fma_define_verify_probe(xp, p);
      }
    }
  }
#endif

  /* Figure out how many nodes are willing to verify */
  verifiers = 0;
  my_id = -1;
  first_verifier = NULL;
  next_verifier = NULL;
  for (h=0; h<fp->num_hosts; ++h) {

    /* find my index within verifiers */
    if (fp->hosts[h] == A.my_host) {
      my_id = verifiers;
    }

    if (fp->hosts[h]->fma_flags & FMA_FLAG_CAN_VERIFY) {
      
      /*get the first verifier*/
      if (first_verifier == NULL){
	first_verifier = fp->hosts [h];
      }
      
      /*get the verifier after myself, if there is one*/
      if (next_verifier == NULL && my_id != -1 && fp->hosts [h] != A.my_host){
	next_verifier = fp->hosts [h];
      }
      ++verifiers;
    }
  }

  /*if there wasn't a next verifier, then wrap*/
  if (next_verifier == NULL)
    next_verifier = first_verifier;
  
  if (A.debug) fma_log("%d verifiers, i am # %d", verifiers, my_id);

  /*define probe for next verifier host. we do this to make sure that
    all of the verifiers are still alive to do work.*/
  
  if (next_verifier != NULL && next_verifier != A.my_host) {
    if (FMA_HOST(next_verifier)->quick_nic != NULL) {
      fma_define_host_probe (next_verifier);  
    }
  }
  

  /*XXX also define host probes for non verifiers, if we decide to do it this way.*/

#if 0
  /* Now, take every Nth link and probe it, starting with M
   * where N is the number of hosts, and M is our verifier index
   */
  l = my_id;
  while (l < FMA_FABRIC(fp)->num_links) {

    if (fma_is_link_xbar_xbar(l)) {

      fma_define_verify_probe(l);
    }

    /* skip to the next link for this host */
    l += verifiers;
  }
#endif

  /*
   * Start probes on every Nth empty xbar port where N is our verifier index
   */
  mod = 0;
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;
    int p;

    xp = fp->xbars[x];
    for (p=0; p<xp->num_ports; ++p) {
      if (xp->topo_ports[p] == NULL) {
	if (mod == my_id) {
	  fma_define_empty_port_probe(xp, p);
	}
	mod = (mod+1) % verifiers;
      }
    }
  }
}

#if 0
/*
 * Test whether a given topo link is xbar-xbar or nor
 */
static int
fma_is_link_xbar_xbar(
  int topo_index)
{
  struct lf_fabric *fp;
  struct lf_xbar *xp1;
  struct lf_xbar *xp2;
  int p1;

  fp = A.fabric;
  
  /* Get pointer to first end of the link */
  xp1 = LF_XBAR(FMA_FABRIC(fp)->link_end_node[topo_index]);
  p1 = FMA_FABRIC(fp)->link_end_port[topo_index];

  /* Make this end is an xbar */
  if (xp1->ln_type != LF_NODE_XBAR) {
    return FALSE;
  }

  /* get the other end */
  xp2 = LF_XBAR(xp1->topo_ports[p1]);

  /* Make sure it is an xbar */
  if (xp2->ln_type != LF_NODE_XBAR) {
    return FALSE;
  }

  return TRUE;
}
#endif

/*
 * Copy clos_level values and link verify assignments from fabric to
 * the topo map.  Assume the xbar arrays are a 1:1 mapping.
 */
static void
fma_copy_fabric_stuff_to_topo()
{
  struct lf_fabric *fp;
  struct lf_topo_xbar *topo_xbar;
  struct lf_topo_link *topo_link;
  struct lf_nic *nicp;
  struct lf_xbar *xp;
  struct lf_verifier *vp;
  int num_links;
  int which;
  int port;
  int vni;
  int ni;
  int l;
  int x;

  fp = A.fabric;
  topo_xbar = LF_TOPO_XBAR_ARRAY(A.map_info->current_topo_map);
  topo_link = LF_TOPO_LINK_ARRAY(A.map_info->current_topo_map);
  num_links = ntohl(A.map_info->current_topo_map->num_links_32);

  /*
   * Copy over all the xbar clos levels
   */
  for (x=0; x<fp->num_xbars; ++x) {
    topo_xbar[x].clos_level_8 = fp->xbars[x]->clos_level;
  }

  /*
   * Copy over link verifier information
   */
  for (l=0; l<num_links; ++l) {

    /* Get to one end of link */
    ni = ntohl(topo_link[l].node_index_32[0]);

    /* If an xbar, check for a verifier */
    if (LF_TOPO_NI_XBAR(ni)) {
      x = LF_TOPO_NI_INDEX(ni);
      port = LF_TOPO_NI_PORT(ni);
      xp = fp->xbars[x];
      which = 0;

      /*
       * If verifier assigned, find the NIC it refers to and
       * save its map index in the link struct
       */
      vp = xp->verifiers + port;

      /* If the other end of the link is the designated verifier,
       * switch pointers to that xbar
       */
      if (vp->ver_nicp == LF_VERIFY_OTHER_END) {
	struct lf_xbar *xp2;
	xp2 = LF_XBAR(xp->topo_ports[port]);
	vp = xp2->verifiers + xp->topo_rports[port];
	which = 1;
      }

      nicp = vp->ver_nicp;
      if (nicp != NULL) {
	vni = LF_TOPO_VERIFIER_INDEX(nicp->n_topo_index,
	    				which, vp->ver_port);
	topo_link[l].verifier_32 = htonl(vni);
      }
    }
  }
}
