/*
 * Routines for dealing with fabric maps on the FMAs
 */
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_fma_comm.h"
#include "lf_topo_map.h"
#include "lf_channel.h"
#include "lf_scheduler.h"

#include "fms.h"
#include "fms_error.h"
#include "fms_fabric.h"
#include "fms_notify.h"
#include "fms_fma.h"
#include "fms_fma_map.h"
#include "fms_state.h"

/*
 * Create the topological representation of the fabric in a form suitable
 * for sending
 */
void
fms_create_topo_map(
  struct lf_fabric *fp)
{
  int topo_map_size;
  struct lf_topo_map *topo_map;

  /* free any previously existing topo map */
  LF_FREE(FMS_FABRIC(fp)->topo_map);

  /* If no hosts or xbars, then no need to proceed with a map */
  if (fp->num_hosts == 0 || fp->num_xbars == 0) {
    return;
  }


  /* create the map */
  topo_map = lf_create_topo_map(fp, &topo_map_size);
  FMS_FABRIC(fp)->topo_map = topo_map;
  if (topo_map == NULL) {
    LF_ERROR(("Error creating topo_map"));
  }

  /*
   * Copy over and fill in some extras
   */
  FMS_FABRIC(fp)->topo_map_size = topo_map_size;
  fms_notify(FMS_EVENT_DEBUG, "topo_map_size = %d", topo_map_size);

  /* fill in mapper info */
  LF_MAC_COPY(topo_map->mapper_mac_addr, FMS_MAPPER_MAC);
  ++F.fabvars->map_version;
  topo_map->map_id_32 = htonl(F.fabvars->map_version);

  return;

 except:
  fms_perror_exit(1);
}

/*
 * update link state in the topo map
 */
void
fms_update_topo_link_state(
  union lf_node *np,
  int port)
{
  struct lf_fabric *fp;
  struct lf_topo_map *topo_map;
  lf_topo_link_state_t *lsp;
  int topo_index;

  fp = F.fabvars->fabric;
  topo_map = FMS_FABRIC(fp)->topo_map;

  if (topo_map != NULL) {
    topo_index = lf_get_node_link_topo_index(np, port);
    lsp = LF_TOPO_LINK_STATE_ARRAY(topo_map);
    lsp[topo_index] = lf_get_node_link_state(np, port);
  }
}

/*
 * Send the latest map to all registered FMAs that need it
 */
int
fms_send_topo_map_to_all_needy_fma(
  struct lf_fabric *fp)
{
  int h;

  /* There should be a map */
  if (FMS_FABRIC(fp)->topo_map == NULL) {
    LF_ERROR(("No map at push time!"));
  }

  fms_notify(FMS_EVENT_INFO, "Sending map to FMAs that need it");

  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL) {
      int rc;

      if (adp->needs_map) {
	fms_notify(FMS_EVENT_DEBUG, "Sending topo map to %s", 
	    fp->hosts[h]->hostname);
	rc = fms_send_topo_map(F.fabvars->fabric, adp);
	if (rc != 0) LF_ERROR(("Error sending topo map to %s", 
	    fp->hosts[h]->hostname));

      } else if (adp->needs_link_state) {
	rc = fms_send_topo_link_state(F.fabvars->fabric, adp);
	if (rc != 0) LF_ERROR(("Error sending link state to %s", 
	    fp->hosts[h]->hostname));
      }
    }
  }

  /* update state machine that map has been sent */
  fms_state_map_sent();

  return 0;

 except:
  return -1;
}

/*
 * Mark every FMA as needing a map update
 */
void
fms_fma_invalidate_map()
{
  struct lf_fabric *fp;
  int h;

  fp = F.fabvars->fabric;

  fms_notify(FMS_EVENT_DEBUG, "All hosts need new map.");

  /* free the existing map */
  LF_FREE(FMS_FABRIC(fp)->topo_map);

  /* mark every host as needing a map */
  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL) {
      adp->needs_map = TRUE;
    }
  }

  return;
}

/*
 * Mark every FMA that doesn't need a map as needing a link state update
 */
void
fms_fma_invalidate_link_state(
  struct lf_fabric *fp)
{
  int h;

  fms_notify(FMS_EVENT_DEBUG, "All hosts need new link state.");

  /* mark every host as needing a map */
  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL && !adp->needs_map) {
      adp->needs_link_state = TRUE;
    }
  }
  return;
}

/*
 * Mark this FMA as needing a map, and schedule a map push
 */
void
fms_fma_needs_topo_map(
  struct fms_fma_desc *adp)
{
  struct lf_fabric *fp;

  fp = F.fabvars->fabric;

  adp->needs_map = TRUE;	/* mark this FMA as needing a map */

  /* Arrange for a map to be sent */
  fms_schedule_map_push();
}

/*
 * Is a map request pending?
 */
int
fms_map_request_pending()
{
  struct lf_fabric *fp;
  struct fms_fabric *fmsp;

  fp = F.fabvars->fabric;
  fmsp = FMS_FABRIC(fp);

  return (fmsp->request_map_timeout != NULL 
          || fmsp->request_map_task != NULL);
}

void
fms_schedule_map_request()
{
  struct lf_fabric *fp;
  struct fms_fabric *fmsp;

  fp = F.fabvars->fabric;
  fmsp = FMS_FABRIC(fp);

  /* If a map request is not pending and not already scheduled, schedule now */
  if (!fms_map_request_pending()) {
    fmsp->request_map_task = lf_schedule_event(fms_fma_request_map,
	                                        NULL, 5*1000);
  }
}

/*
 * Send the topological part of the map to an FMA
 */
int
fms_send_topo_map(
  struct lf_fabric *fp,
  struct fms_fma_desc *adp)
{
  int rc;

  /* send topo map message header */
  rc = fms_fma_write(adp, LF_FMA_TOPO_MAP,
                     FMS_FABRIC(fp)->topo_map, FMS_FABRIC(fp)->topo_map_size);
  if (rc == -1) {
    LF_ERROR(("Error sending topo_map header to %s", adp->hostname));
  }

  adp->needs_map = FALSE;	/* map no longer needed */
  adp->needs_link_state = FALSE;
  
  return 0;

 except:
  return -1;
}

/*
 * Send the topological link state to an FMA
 */
int
fms_send_topo_link_state(
  struct lf_fabric *fp,
  struct fms_fma_desc *adp)
{
  int rc;
  int size;

  /* send topo link state message header */
  size = LF_TOPO_LINK_STATE_SIZE(FMS_FABRIC(fp)->topo_map);
  rc = fms_fma_write(adp, LF_FMA_TOPO_LINK_STATE,
       		LF_TOPO_LINK_STATE_ARRAY(FMS_FABRIC(fp)->topo_map), size);
  if (rc == -1) {
    LF_ERROR(("Sending link state to %s", adp->hostname));
  }

  adp->needs_link_state = FALSE;	/* link state no longer needed */
  
  return 0;

 except:
  return -1;
}
