/*
 * A shared routine to assign link verify responsibilities to
 * various hosts in the fabric
 */

#include <stdio.h>

#include "libfma.h"
#include "lf_internal.h"
#include "lf_fabric.h"
#include "lf_fma_flags.h"

/*
 * local prototypes
 */
static int lf_alv_alloc_space(struct lf_fabric *fp);
static void lf_alv_free_space(struct lf_fabric *fp);
static struct lf_nic *lf_alv_closest_nic(struct lf_xbar *start_xp,
  int skip_port, int sequence, int *distp, int *loadp, int *portp);

struct lf_alv_xbar {
  void *old_user;
  int best_load[LF_MAX_XBAR_PORTS];
  int min_dist[LF_MAX_XBAR_PORTS];

  int sequence;
  struct lf_xbar *next;
};
#define LV_XBAR(X) ((struct lf_alv_xbar *)((X)->user.v))

struct lf_alv_nic {
  void *old_user;
  int alv_load;
};
#define LV_NIC(N) ((struct lf_alv_nic *)((N)->user.v))

int
lf_assign_link_verifies(
  struct lf_fabric *fp)
{
  struct lf_xbar *xp;
  struct lf_xbar *xp2;
  int port2;
  union lf_node *np;
  struct lf_nic *nicp;
  int nic_port;
  struct lf_verifier *vp;
  struct lf_verifier *vp2;
  int sequence;		/* used as "seen" in BFS */
  int dist;
  int load;
  int rc;
  int p;
  int x;

  sequence = 0;

  /* Allocate all local structs */
  rc = lf_alv_alloc_space(fp);
  if (rc == -1) LF_ERROR(("Error allocating link verify private space."));

  /*
   * Clear out xbar verifiers
   */
  for (x=0; x<fp->num_xbars; ++x) {
    xp = fp->xbars[x];
    for (p=0; p<xp->num_ports; ++p) {
      xp->verifiers[p].ver_nicp = NULL;
    }
  }

  /*
   * Find a verifier for each link
   */
  for (x=0; x<fp->num_xbars; ++x) {
    xp = fp->xbars[x];

    for (p=0; p<xp->num_ports; ++p) {

      /* Where are we connected */
      np = xp->topo_ports[p];

      /* skip if no conn or conn is NIC */
      if (np == NULL || np->ln_type != LF_NODE_XBAR) {
	continue;
      }
      
      /*
       * find closest, least loaded NIC not through this port.
       * We may already have a verifier, found while inspecting the 
       * xbar on the other end of this link.  If ours is better than
       * his, replace it!
       */
      nicp = lf_alv_closest_nic(xp, p, ++sequence, &dist, &load, &nic_port);

      /* If no NICs to be found, skip it */
      if (nicp == NULL) continue;

      xp2 = LF_XBAR(np);
      port2 = xp->topo_rports[p];
      vp = &xp->verifiers[p];
      vp2 = &xp2->verifiers[port2];

      /* If our distance is better or dist is same and load better, use this */
      if (vp->ver_nicp == NULL
	  || dist < LV_XBAR(xp)->min_dist[p]
	  || (dist == LV_XBAR(xp)->min_dist[p]
	      && load < LV_XBAR(xp)->best_load[p])) {


	/* If there was a previous verifier, decrement his load */
	if (vp2->ver_nicp != NULL) {
	  --LV_NIC(vp2->ver_nicp)->alv_load;
	}

	/* Fill in both ends of link */
	vp->ver_nicp = nicp;
	vp->ver_port = nic_port;
	LV_XBAR(xp)->min_dist[p] = dist;
	LV_XBAR(xp)->best_load[p] = load;

	/* beware loopback! */
	if (vp != vp2) {
	  vp2->ver_nicp = LF_VERIFY_OTHER_END;
	  vp2->ver_port = 0;
	  LV_XBAR(xp2)->min_dist[port2] = dist;
	  LV_XBAR(xp2)->best_load[port2] = load;
	}

	++LV_NIC(nicp)->alv_load;
      }
    }
  }

  rc = 0;

 except:
  /* return all xbar and host user structs to original values */
  lf_alv_free_space(fp);
  return rc;
}


/*
 * Find the closest downhill NIC with the lowest load
 * "Downhill" means we never visit xbars with a higher clos level
 */
static struct lf_nic *
lf_alv_closest_nic(
  struct lf_xbar *start_xp,
  int skip_port,
  int sequence,
  int *distp,
  int *loadp,
  int *portp)
{
  struct lf_xbar *work_list;
  struct lf_xbar *new_list;
  struct lf_xbar *xp;
  struct lf_xbar *xp2;
  union lf_node *np2;
  struct lf_nic *nicp;
  struct lf_nic *best_nicp;
  int dist;
  int best_load;
  int best_port;
  int p;

  best_load = 999999;
  best_nicp = NULL;
  dist = 0;

  LV_XBAR(start_xp)->next = NULL;
  new_list = start_xp;
  work_list = NULL;
  LV_XBAR(start_xp)->sequence = sequence;		/* mark this as seen */

  while (new_list != NULL) {
    work_list = new_list;
    new_list = NULL;

    while (work_list != NULL) {
      xp = work_list;
      work_list = LV_XBAR(xp)->next;


      for (p=0; p<xp->num_ports; ++p) {

	/* don't go out the port we are ultimately testing */
	if (xp == start_xp && p == skip_port) continue;

	np2 = xp->topo_ports[p];

	/* skip if not connected */
	if (np2 == NULL) continue;

	/* If a NIC found, see if this is the best load */
	if (np2->ln_type == LF_NODE_NIC) {
	  nicp = LF_NIC(np2);

	  /* see if we have a new winner */
	  if ((nicp->host->fma_flags & FMA_FLAG_CAN_VERIFY)
	      && LV_NIC(nicp)->alv_load < best_load) {
	    best_load = LV_NIC(nicp)->alv_load;
	    best_nicp = nicp;
	    best_port = xp->topo_rports[p];
	  }

	} else if (np2->ln_type == LF_NODE_XBAR) {
	  xp2 = LF_XBAR(np2);

	  /* possibly skip for various reasons */
	  if (LV_XBAR(xp2)->sequence == sequence	/* already seen */
	      || xp2->clos_level > xp->clos_level) {	/* uphill */
	    continue;
	  }

	  /* Add this xbar to the new work list for next time around */
	  LV_XBAR(xp2)->next = new_list;
	  new_list = xp2;
	  LV_XBAR(xp2)->sequence = sequence;		/* mark this as seen */
	}
      }
    }

    /* If a host found this time through, we're done */
    if (best_nicp != NULL) {
      *distp = dist;
      *loadp = best_load;
      *portp = best_port;
      return best_nicp;
    }

    ++dist;
  }

  return NULL;		/* no joy in Mudville */
}

/*
 * Alloacate space for working variables
 */
static int
lf_alv_alloc_space(
  struct lf_fabric *fp)
{
  struct lf_alv_xbar *alvxp;
  struct lf_alv_nic *alvnicp;
  struct lf_xbar *xp;
  struct lf_nic *nicp;
  struct lf_host *hp;
  int num_nics;
  int x;
  int h;
  int ni;
  int n;

  alvxp = NULL;
  alvnicp = NULL;

  /*
   * Assign all xbar structs, if any xbars
   */
  if (fp->num_xbars > 0) {
    /* allocate data space (all are in one alloced array for efficiency) */
    LF_CALLOC(alvxp, struct lf_alv_xbar, fp->num_xbars);

    for (x=0; x<fp->num_xbars; ++x) {
      xp = fp->xbars[x];
      alvxp[x].old_user = xp->user.v;
      xp->user.v = alvxp + x;
    }
  }

  /* first, count all the nics */
  num_nics = 0;
  for (h=0; h<fp->num_hosts; ++h) {
    hp = fp->hosts[h];
    num_nics += hp->num_nics;
  }
    
  /* If any NICs found, assign all NIC structs */
  if (num_nics > 0) {
    LF_CALLOC(alvnicp, struct lf_alv_nic, num_nics);

    ni=0;
    for (h=0; h<fp->num_hosts; ++h) {
      hp = fp->hosts[h];

      for (n=0; n<hp->num_nics; ++n) {
	nicp = hp->nics[n];
	alvnicp[ni].old_user = nicp->user.v;
	nicp->user.v = alvnicp + ni;
	++ni;
      }
    }
  }

  return 0;

 except:
  LF_FREE(alvxp);
  LF_FREE(alvnicp);
  return -1;
}

/*
 * Free out temporary working space
 */
static void
lf_alv_free_space(
  struct lf_fabric *fp)
{
  struct lf_alv_xbar *alvxp;
  struct lf_alv_nic *alvnicp;
  struct lf_xbar *xp;
  struct lf_host *hp;
  struct lf_nic *nicp;
  int ni;
  int n;
  int x;
  int h;

  /* Free all xbar structs */
  if (fp->num_xbars > 0) {
    alvxp = fp->xbars[0]->user.v;
    for (x=0; x<fp->num_xbars; ++x) {
      xp = fp->xbars[x];
      xp->user.v = alvxp[x].old_user;
    }
    LF_FREE(alvxp);
  }

  /* Free NIC structs */
  alvnicp = NULL;
  ni = 0;
  for (h=0; h<fp->num_hosts; ++h) {
    hp = fp->hosts[h];

    if (hp->num_nics > 0) {

      /* The first NIC is the start of our allocated struct */
      if (alvnicp == NULL) {
	alvnicp = hp->nics[0]->user.v;
      }

      for (n=0; n<hp->num_nics; ++n) {
	nicp = hp->nics[n];
	nicp->user.v = alvnicp[ni].old_user;
	++ni;
      }
    }
  }
  LF_FREE(alvnicp);
}
