/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

static const char __idstring[] = "@(#)$Id: mx__debug_dump.c,v 1.30.4.2 2006/04/11 23:32:37 loic Exp $";
 
/*
   Contains debugging functions to dump the state of the lib.
*/

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#  include "mx__partner.h"
#include "mx__endpoint.h"
#include "mx__shim.h"
#include "mx__debug_dump.h"
#include "mx_byteswap.h"
#include "mx__handle_map.h"
#include "mx__lib.h"
#include "mx__request.h"

#if ! MX_OS_WINNT && ! defined MX_KERNEL
#include <signal.h>

static void (*old_signal)(int);
static int signal_installed;

static void
handle_signal(int sig)
{
  struct mx_endpoint *ep;
  mx_printf("Signal %d received\n", sig);

  for (ep = Mx_endpoints; ep; ep = ep->next) {
    mx_printf("Endpoint %d\n", ep->addr.endpoint_id);
    mx__dump_endpoint(ep, 0);
  }
  for (ep = Mx_endpoints; ep; ep = ep->next) {
    mx_printf("Endpoint %d/Request checks\n", ep->addr.endpoint_id);
    mx__conservation_of_matter(ep);
  }
}


#endif

static const char *rtype_names[] = {
  "send_null",
  "send_small",
  "send_medium",
  "send_large",
  "recv",
  "recv_large",
  "recv_self",
  "recv_shm",
  "connect",
  "connect_reply",
};

static const char *rstate_names[] = {
  "send_queued",
  "pending",
  "buffered",
  "completed",
  "dead",
  "acked",
  "mcp-posted",
  "notified",
  NULL
};

void
mx__dump_request(struct mx_endpoint *ep, union mx_request *r)
{
  int i;
  struct mx__partner *partner;

  mx_printf("\ttype: %d (%s)\n", r->basic.type, rtype_names[r->basic.type]);
  mx_printf("\tstate (0x%x):\n",r->basic.state);
  for (i=0;rstate_names[i];i++) {
    if (r->basic.state & (1<< i))
      mx_printf("\t\t%s\n",rstate_names[i]);
  }
  mx_printf("\tmcp_handle : %d\n", r->basic.mcp_handle);
  switch (r->basic.type) {
  case MX__REQUEST_TYPE_RECV:
  case MX__REQUEST_TYPE_RECV_LARGE:
  case MX__REQUEST_TYPE_RECV_SHM:
  case MX__REQUEST_TYPE_RECV_SELF:
#ifndef MX_KERNEL    
    mx_printf("\tseg:%p,%d\n", r->recv.segments[0].segment_ptr, r->recv.segments[0].segment_length);
#endif
    if (!r->recv.unexpected) {
      mx_printf("\tmatch_mask:(0x%08x_%08x)\n",
		MX_U32(r->recv.match_mask), MX_L32(r->recv.match_mask));
      mx_printf("\tmatch_val:(0x%08x_%08x)\n",
		MX_U32(r->recv.match_info), MX_L32(r->recv.match_info));
    }
    mx_printf("\tmatched_val:(0x%08x_%08x)\n",
	      MX_U32(r->recv.basic.status.match_info), MX_L32(r->recv.basic.status.match_info));
    mx_printf("\tslength=%d,rlength=%d,accum_len=%d\n", r->recv.basic.status.msg_length, r->recv.r_length, r->recv.accum);
    mx_printf("\tunexpected=%d\n", r->recv.unexpected);
    if (r->basic.type == MX__REQUEST_TYPE_RECV_LARGE) {
      mx_printf("\tlocal_rdma_id=%d\n", r->recv.local_rdma_id);
      mx_printf("\tremote_rdma=%d\n", r->recv.remote_rdma);
    }
    break;
  case MX__REQUEST_TYPE_SEND_NULL:
  case MX__REQUEST_TYPE_SEND_SMALL:
  case MX__REQUEST_TYPE_SEND_MEDIUM:
  case MX__REQUEST_TYPE_SEND_LARGE:
#ifndef MX_KERNEL
    mx_printf("\tseg:%p,%d\n", r->send.segments[0].segment_ptr, r->send.segments[0].segment_length);
#endif
    mx_printf("\tdest:peer_index=%d,eid=%d,seqnum=%d\n", 
	      (int)htons(r->send.dest_address.peer_index_n), 
	      r->send.dest_address.endpoint_id, r->send.send_seq);
    mx_printf("\tslength=%d,xfer_length=%d\n", r->send.length, r->send.basic.status.xfer_length);
    if (r->basic.type == MX__REQUEST_TYPE_SEND_LARGE) {
      mx_printf("\tlocal_rdma_id=%d\n", r->recv.local_rdma_id);
    }
    mx_printf("\tcaller: %p\n", r->send.caller);
    break;
  case MX__REQUEST_TYPE_CONNECT:
  case MX__REQUEST_TYPE_CONNECT_REPLY:
    mx_printf("\txid=0x%x,peer_index=0x%x,peer_eid=0x%x\n", r->connect.connect_seqnum_n, 
	      htons(r->connect.peer_index_n), r->connect.peer_endpoint_id);
    partner = mx__endpoint_lookup_partner(ep, r->connect.peer_endpoint_id, ntohs(r->connect.peer_index_n));
    if (partner) {
      char s[18];
      char hostname[MX_MAX_HOSTNAME_LEN];
      mx_printf("\tpeer_mac=%s", mx__nic_id_to_str(s, partner->nic_id, 18));
      hostname[0] = 0;
      mx_nic_id_to_hostname(partner->nic_id, hostname);
      mx_printf("(%s)\n", hostname);
    }
    break;
  default:
    mx_printf("\tunknown request type:%d\n", r->basic.type);
  }
  mx_printf("\n");
}

static void
mx__dump_request_list(struct mx_endpoint *ep, union mx_request *list, const char *list_name, int fulldump)
{
  union mx_request *r;
  int i;
  mx_printf("%s: %d items\n", list_name, mx__count_requests(list));
  for (i = 0, r = list->basic.next; r != list && (fulldump || i < 6); r= r->basic.next,i++) {
    mx__dump_request(ep, r);
  }
}

void
mx__dump_endpoint(struct mx_endpoint *ep, int fulldump)
{
  unsigned int i;
  mx_printf("handle_map: total=%d,free=%d,first=%d\n", 
	    ep->handle_map->total_count,
	    ep->handle_map->free_count,
	    ep->handle_map->first_free);
  mx_printf("eventq: count=%d,index=%d,flow=%d\n", ep->event_count, ep->eventq_index, ep->eventq_flow);
  mx_printf("rdmas: max=%d, busy=%d\n", ep->rdmas.max, ep->rdmas.count);
  mx_printf("req-lookaside: free=%d,busy=%d,allocated=%d\n", ep->req_lookaside.count, 
	    ep->req_lookaside.alloc_count - ep->req_lookaside.count, 
	    ep->req_lookaside.alloc_count);
  for (i=0;i<ep->max_peers * ep->max_endpoints;i++) {
    struct mx__partner *p = ep->remote_ep[i];
    int app_rank = ep->application_rank_info ? ep->application_rank_info[i] : -1;
    if (p && (p->send_seq || p->recv_seq || app_rank != -1)) {
      struct mx__early *e;
      mx_printf("peer %d: app_rank=%d, send_seq =%d, recv_seq = %d\n", 
	      i, app_rank, p->send_seq, p->recv_seq);
      for (e= p->early.next; e != &p->early; e= e->next) {
	mx_printf("early seq=%d, length = %d, type = %d \n",
		  e->msg_seq, e->data_length, e->type);
      }
    }
  }
  mx_printf("Send-shm-pending:%d\n", ep->sendshm_count);
  mx_printf("Send-self-pending:%d\n", ep->sendself_count);
  mx__dump_request_list(ep, &ep->send_reqq, "Send", fulldump);
  mx__dump_request_list(ep, &ep->recv_reqq, "Recv", fulldump);
  mx__dump_request_list(ep, &ep->doneq, "Done", fulldump);
  mx__dump_request_list(ep, &ep->unexpq, "Unexp", fulldump);
  mx__dump_request_list(ep, &ep->buffered_sendq, "Buffered send", fulldump);
  mx__dump_request_list(ep, &ep->connectq, "Connect", fulldump);
  mx__dump_request_list(ep, &ep->connect_replyq, "Connect replies", fulldump);
  mx__dump_request_list(ep, &ep->multifrag_recvq, "Multifraq recvs", fulldump);
  mx__dump_request_list(ep, &ep->large_sendq, "Large send", fulldump);
  mx__dump_request_list(ep, &ep->large_getq, "Large gets", fulldump);
}


void
mx__set_app_rank(struct mx_endpoint *ep, mx_endpoint_addr_t *addr, int n)
{
  mx__endpoint_addr_t *addrp = (mx__endpoint_addr_t*)addr;
  
  if (!ep->application_rank_info) {
    int nbytes = sizeof(ep->application_rank_info[0])*ep->max_endpoints * ep->max_peers;
    ep->application_rank_info = mx_malloc(nbytes);
    if (!ep->application_rank_info) {
      mx_printf("Warning: mx_set_app_rank failed: NO MEM\n");
      return;
    }
    memset(ep->application_rank_info, -1, nbytes);
  }
  ep->application_rank_info[ntohs(addrp->peer_index_n)*ep->max_endpoints + addrp->endpoint_id] = n;
}


void
mx__init_signal_handling(void)
{
#if ! MX_OS_WINNT && ! defined MX_KERNEL
  if (mx__opt.sigusr1_handler) {
    old_signal = signal(SIGUSR1,handle_signal);
    signal_installed = 1;
  }
#endif
}


void
mx__finalize_signal_handling(void)
{
#if ! MX_OS_WINNT && ! defined MX_KERNEL
  if (signal_installed) {
    signal(SIGUSR1,old_signal);
  }
#endif
}
