/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_auto_config.h"
#include "myriexpress.h"
#include "mx__lib_types.h"
#include "mx__sleep.h"
#include "mx__mcp_request_ring.h"
#  include "mx_stbar.h"
#  include "mx_byteswap.h"
#include "mx__lib.h"
#include "mx__requests.h"
#include "mx__request.h"
#include "mx__shmem.h"
#include "mx__driver_interface.h"
#include "mx__handle_map.h"
#include "mx__partner.h"
#include "mx__endpoint.h"

void mx__uwake_req(struct mx_endpoint *ep)
{
#if MX_USE_SHMEM
  uint32_t type, read_idx;
#endif
  if (ep->wake_pending == 0) {
    mcp_ureq_t *req_user;

    ep->wake_mcp_handle = mx__endpoint_alloc_mcp_handle(ep);
    req_user = mx__get_ureq(ep->req_ring, ep->wake_mcp_handle);
    mx__post_ureq_wake(req_user, ep->eventq_flow);
    ep->wake_pending = 1;
  }
#if MX_USE_SHMEM
  if (mx__opt.disable_shmem)
    return;
  ep->shm->shmq->waiting = 1;
  MX_READBAR();
  read_idx = ep->shm->shmq->read_idx;
  type = ep->shm->shmq->queue[MX__SHM_REQ_SLOT(read_idx)].type;
  if (!((type ^ read_idx) & MX__SHM_REQ_CNT)) {
    /* rather than trying to avoid a race (between sleeping and event
       arriving in shm queue), it is easier to detect it here, and do
       as if it didn't happen (act like if the sender actually arrived
       a few nanoseconds later, and do the wake for him */
    if (ep->app_waiting)
      mx__app_wake(ep->handle);
    mx__wake(ep->handle);
  }
#endif
}

void mx__sleep_on_request(struct mx_endpoint *endpoint, union mx_request *r, uint32_t timeout, int *waiters, union mx_request *list)
{
  struct timeval start,cur;
  int stimeout = timeout * 1000;
  mx_wait_t x;
  int ret;

  r->basic.wait_state = MX__WAIT_STATE_WAITING;
  if (list) {
    mx__enqueue_request(list, r);
  }
  waiters[0]++;
  MX__EVENT_INIT(&r->basic.event);
  if (MX_THREAD_SAFE && !mx__opt.monothread) {
    if (endpoint->app_waiting) {
      mx__uwake_req(endpoint);
      if (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
	if (timeout == MX_INFINITE) {
	  MX__EVENT_WAIT(&r->basic.event, &endpoint->lock);
	}
	else {
	  MX__EVENT_WAIT_TIMEOUT(&r->basic.event, &endpoint->lock, timeout);
	}
      } 
    } else {
      start.tv_sec = 0;
      start.tv_usec = 0;

      while (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
	if (timeout == MX_INFINITE) {
	  x.timeout = MX_MAX_WAIT;
	} else {
	  mx_gettimeofday(&cur, NULL);
	  if (!start.tv_sec)
	    start = cur;
	  stimeout -= (cur.tv_sec - start.tv_sec) * 1000000 + (cur.tv_usec - start.tv_usec);
	  x.timeout = stimeout / 1000;
	  if ((int)x.timeout <= 0)
	    break;
	}
	mx__uwake_req(endpoint);
	x.status = 0;
	x.mcp_wake_events = 0;
	endpoint->app_waiting = 1;
	MX__MUTEX_UNLOCK(&endpoint->lock);
	ret = mx__app_wait(endpoint->handle, &x);
	if (ret != 0) {
	  mx_fatal("mx__app_wait failed, check kernel logs for error messages");
	}
	MX__MUTEX_LOCK(&endpoint->lock);
	endpoint->app_waiting = 0;
	mx_assert(endpoint->wake_pending >= x.mcp_wake_events);
	if (x.mcp_wake_events) {
	  endpoint->wake_pending -= x.mcp_wake_events;
	  mx__endpoint_free_mcp_handle(endpoint,endpoint->wake_mcp_handle);
	}
	mx__luigi(endpoint);
	/* If another thread is blocked, submit a wake request
	   so the progression thread will be awoken in a timely
	   fashion to wake the blocked thread */
	if (r->basic.wait_state != MX__WAIT_STATE_WAITING &&
	    ((endpoint->waiters > 0 || 
	      endpoint->peek_waiters > 0 || endpoint->probe_waiters > 0)))
	  mx__uwake_req(endpoint);
      }
    }
  } else {
#ifndef MX_KERNEL
    if (timeout == MX_INFINITE) {
      while (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
	mx__luigi(endpoint);
	mx_sched_yield();
      }
    } else {
      start.tv_sec = 0;
      start.tv_usec = 0;
      while (r->basic.wait_state == MX__WAIT_STATE_WAITING && stimeout > 0) {
	mx_gettimeofday(&cur, NULL);
	if (!start.tv_sec)
	  start = cur;
	mx_sched_yield();
	stimeout -= (cur.tv_sec - start.tv_sec) * 1000000 + (cur.tv_usec - start.tv_usec);
	start = cur;
	mx__luigi(endpoint);
      }
    }
#else
    mx_fatal("kernel always has a thread");
#endif
  }
  MX__EVENT_DESTROY(&r->basic.event);
  if (r->basic.wait_state == MX__WAIT_STATE_WAITING) {
    r->basic.wait_state = MX__WAIT_STATE_COMPLETE;
    waiters[0]--;
    if (list) {
      mx__spliceout_request(list, r);
    }
  }
}
